from __future__ import annotations

from numbers import Integral
from typing import Optional, Tuple

import numpy as np
from chex import Array
from jax import numpy as jnp
from jax.lax import scan

from .dataclass import dataclass
from .dtypes import real_dtype
from .generator import Generator

__all__ = ['leaky_integrate', 'diffused_leaky_integrate', 'leaky_data_weight',
           'leaky_integrate_time_series', 'leaky_covariance']


def leaky_integrate(value: Array,
                    time_step: Array,
                    drift: Optional[Array] = None,
                    decay: Optional[Array] = None,
                    *,
                    leaky_average: bool = False) -> Array:
    """
    Update the value so that it is the leaky integral (or leaky average).
    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        decay: If provided, must be positive, and the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    if drift is None:
        if decay is None:
            return value
        return value * jnp.exp(-decay * time_step)

    if decay is None:
        if leaky_average:
            raise ValueError
        return value + drift * time_step

    scaled_integrand = (drift / decay) * -jnp.expm1(-decay * time_step)

    if leaky_average:
        scaled_integrand *= decay.real

    return value * jnp.exp(-decay * time_step) + scaled_integrand


def diffused_leaky_integrate(value: Array,
                             time_step: Array,
                             rng: Generator,
                             diffusion: Array,
                             drift: Optional[Array] = None,
                             decay: Optional[Array] = None,
                             *,
                             leaky_average: bool = False) -> Tuple[Array, Generator]:
    """
    Update an Ornstein-Uhlenbeck process.

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        rng: The random number generator for the stochastic process.
        diffusion: The diffusion for the stochastic process.
        decay: If provided, the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    variance = (diffusion * time_step
                if decay is None
                else diffusion / decay * -jnp.expm1(-decay * time_step))
    jump, new_rng = rng.normal(jnp.sqrt(variance), shape=value.shape)
    return (leaky_integrate(value, time_step, drift, decay, leaky_average=leaky_average) + jump,
            new_rng)


def leaky_data_weight(iterations_times_time_step: Array,
                      decay: Array) -> Array:
    """
    Returns: The amount of data that has been incorporated and has not been decayed.
    """
    return leaky_integrate(0.0, iterations_times_time_step, 1.0, decay, leaky_average=True)


@dataclass
class _FilterCarry:
    iterations: Array
    value: Array


def leaky_integrate_time_series(time_series: Array, decay: Array) -> Array:
    """
    Args:
        time_series: A sequence of
        f: A function that maps from value, drift to new_value.
        reweighted: Rescales the early
    """
    def g(carry: _FilterCarry, drift: Array) -> Tuple[_FilterCarry, Array]:
        new_iterations = carry.iterations + 1.0
        data_weight = leaky_data_weight(new_iterations, decay)

        new_value = leaky_integrate(carry.value, 1.0, drift, decay, leaky_average=True)
        new_carry = _FilterCarry(new_iterations, new_value)
        outputted_value = new_value / data_weight
        return new_carry, outputted_value

    # Cast the dtype from integer to floating point to prevent integer rounding.
    initial_value = jnp.zeros(time_series[0].shape,
                              dtype=(real_dtype
                                     if issubclass(time_series.dtype.type, Integral)
                                     else time_series.dtype))
    initial_carry = _FilterCarry(0.0, initial_value)

    _, filtered_time_series = scan(g, initial_carry, time_series)
    return filtered_time_series


def leaky_covariance(x_time_series: Array,
                     y_time_series: Array,
                     decay: Array,
                     covariance_matrix: bool = False) -> Array:
    if covariance_matrix:
        if x_time_series.shape[0] != y_time_series.shape[0]:
            raise ValueError
        s = (np.newaxis,)

        def times(a: Array, b: Array) -> Array:
            return a[(..., *(s * (b.ndim - 1)))] * b[(slice(None), *(s * (a.ndim - 1)))]
    else:
        if x_time_series.shape != y_time_series.shape:
            raise ValueError
        times = jnp.multiply
    x = leaky_integrate_time_series(x_time_series, decay)
    y = leaky_integrate_time_series(y_time_series, decay)
    xy = leaky_integrate_time_series(times(x_time_series, y_time_series), decay)
    return xy - times(x, y)



import dataclasses
from dataclasses import MISSING, Field, FrozenInstanceError, InitVar, asdict, astuple
from dataclasses import fields as d_fields
from dataclasses import is_dataclass, replace
from functools import partial
from typing import (Any, Callable, Hashable, Iterable, List, Mapping, MutableMapping, Optional,
                    Sequence, Tuple, Type, TypeVar, overload)

from jax.tree_util import register_pytree_node

from .annotations import PyTree
from .display import display_class, display_generic, display_key_and_value
from .testing import get_relative_test_string, get_test_string, jax_allclose

__all__ = ['dataclass', 'field', 'Field', 'FrozenInstanceError', 'InitVar', 'MISSING',
           # Helper functions.
           'fields', 'asdict', 'astuple', 'replace', 'is_dataclass', 'field_names',
           'field_names_and_values', 'field_names_values_metadata', 'field_values',
           # New functions.
           'document_dataclass']


T = TypeVar('T', bound=Any)


@overload
def dataclass(*, init: bool = True, repr_: bool = True, eq: bool = True,
              order: bool = False) -> Callable[
                  [Type[T]], Type[T]]:
    ...


@overload
def dataclass(cls: Type[T], *, init: bool = True, repr_: bool = True, eq: bool = True,
              order: bool = False) -> Type[T]:
    ...


# TODO: use positional-only arguments
def dataclass(cls: Optional[Type[T]] = None, *, init: bool = True, repr_: bool = True,
              eq: bool = True, order: bool = False) -> Any:
    """
    Returns the same class as was passed in, with dunder methods added based on the fields defined
    in the class.

    Examines PEP 526 annotations to determine fields.  Default values for fields are provided using
    assignment.  To mark fields as JAX static fields rather than JAX pytree fields, use the `field`
    function.

    For example::
    ```python
    from __future__ import annotations

    from typing import ClassVar

    from tjax import dataclass, field, Array
    from jax import numpy as jnp
    from jax import grad

    @dataclass
    class LearnedParameter:
        weight: Array
        constrain_positive: bool = field(static=True)
        minimum_positive_weight: ClassVar[Array] = 1e-6

        def trained(self,
                    self_bar: LearnedParameter,
                    learning_rate: float) -> LearnedParameter:
            weight_bar = self_bar.weight
            weight = self.weight - weight_bar * learning_rate
            if self.constrain_positive:
                weight = jnp.maximum(weight, self.minimum_positive_weight)
            return LearnedParameter(weight=weight,
                                    constrain_positive=self.constrain_positive)

    def loss(w: LearnedParameter) -> float:
        return jnp.square(w.weight - 3.3)

    w = LearnedParameter(2.0, True)
    w_bar = grad(loss)(w)
    new_w = w.trained(w_bar, 1e-4)
    ```

    `dataclass` includes a convenient replace method::

        w.replace(weight=3.4)

    Since this dataclass is a pytree, all of JAX's functions that accept pytrees work with it,
    including iteration, differentiation, and `jax.tree_util` functions.

    Another benefit is the display of dataclasses.  `print(new_w)` gives::
    ```
    LearnedParameter
        weight=Jax Array ()
                2.0003
        constrain_positive=True
    ```
    """
    if cls is None:
        return partial(dataclass, init=init, repr_=repr_, eq=eq, order=order)

    # pylint: disable=protected-access

    # Apply dataclass function to cls.
    data_clz: Type[T] = dataclasses.dataclass(cls, init=init, repr=repr_, eq=eq,
                                              order=order, frozen=True)  # type: ignore

    # Partition fields into hashed, tree, and uninitialized.
    static_fields: List[str] = []
    nonstatic_fields: List[str] = []

    for field_info in dataclasses.fields(data_clz):  # type: ignore
        if not field_info.init:
            continue
        if field_info.metadata.get('static', False):
            static_fields.append(field_info.name)
        else:
            nonstatic_fields.append(field_info.name)

    # Generate additional methods.
    def __str__(self: T) -> str:
        return str(self.display())

    def tree_flatten(x: T) -> Tuple[Sequence[PyTree], Hashable]:
        hashed = tuple(getattr(x, name) for name in static_fields)
        trees = tuple(getattr(x, name) for name in nonstatic_fields)
        return trees, hashed

    def tree_unflatten(cls: Type[T], hashed: Hashable, trees: Sequence[PyTree]) -> T:
        if not isinstance(hashed, tuple):
            raise TypeError
        hashed_args = dict(zip(static_fields, hashed))
        tree_args = dict(zip(nonstatic_fields, trees))
        return cls(**hashed_args, **tree_args)

    # Assign methods to the class.
    if data_clz.__str__ is object.__str__:
        data_clz.__str__ = __str__  # type: ignore
    if not hasattr(data_clz, 'display'):
        data_clz.display = display_dataclass  # type: ignore
    data_clz.tree_flatten = tree_flatten  # type: ignore
    data_clz.tree_unflatten = classmethod(tree_unflatten)  # type: ignore
    data_clz.replace = replace  # type: ignore

    # Assign field lists to the class.
    data_clz.nonstatic_fields = nonstatic_fields  # type: ignore
    data_clz.static_fields = static_fields  # type: ignore

    # Register the class as a JAX PyTree.
    register_pytree_node(data_clz, tree_flatten, data_clz.tree_unflatten)  # type: ignore

    # Register the dynamically-dispatched functions.
    display_generic.register(data_clz, data_clz.display)  # type: ignore
    get_test_string.register(data_clz, get_dataclass_test_string)  # type: ignore
    get_relative_test_string.register(data_clz, get_relative_dataclass_test_string)  # type: ignore
    return data_clz


def display_dataclass(value: T, show_values: bool = True, indent: int = 0) -> str:
    retval = display_class(type(value))
    for field_info in dataclasses.fields(value):  # type: ignore
        retval += display_key_and_value(
            field_info.name, getattr(value, field_info.name), "=", show_values, indent)
    return retval


def get_dataclass_test_string(actual: Any, rtol: float, atol: float) -> str:
    retval = f"{type(actual).__name__}("
    retval += ",\n".join(
        f"{fn}=" + get_test_string(getattr(actual, fn), rtol, atol)
        for fn in actual.nonstatic_fields)
    if actual.nonstatic_fields and actual.static_fields:
        retval += ',\n'
    retval += ",\n".join(
        f"{fn}=" + get_test_string(getattr(actual, fn), rtol, atol)
        for fn in actual.static_fields)
    retval += ")"
    return retval


def get_relative_dataclass_test_string(actual: Any,
                                       original_name: str,
                                       original: Any,
                                       rtol: float,
                                       atol: float) -> str:
    retval = f"{original_name}.replace("
    retval += ",\n".join(
        f"{fn}=" + get_relative_test_string(f"{original_name}.{fn}",
                                            getattr(actual, fn),
                                            getattr(original, fn),
                                            rtol,
                                            atol)
        for fn in actual.nonstatic_fields
        if not jax_allclose(getattr(actual, fn), getattr(original, fn), rtol=rtol, atol=atol))
    retval += ")"
    return retval


# NOTE: Actual return type is 'Field[T]', but we want to help type checkers
# to understand the magic that happens at runtime.
# pylint: disable=redefined-builtin
@overload  # `default` and `default_factory` are optional and mutually exclusive.
def field(*, static: bool = False, default: T, init: bool = ..., repr: bool = ...,
          hash: Optional[bool] = ..., compare: bool = ...,
          metadata: Optional[Mapping[str, Any]] = ...) -> T:
    ...


@overload
def field(*, static: bool = False, default_factory: Callable[[], T], init: bool = ...,
          repr: bool = ..., hash: Optional[bool] = ..., compare: bool = ...,
          metadata: Optional[Mapping[str, Any]] = ...) -> T:
    ...


@overload
def field(*, static: bool = False, init: bool = ..., repr: bool = ..., hash: Optional[bool] = ...,
          compare: bool = ..., metadata: Optional[Mapping[str, Any]] = ...) -> Any:
    ...


def field(*, static: bool = False, default: Any = MISSING,
          default_factory: Callable[[], Any] = MISSING, init: bool = True,  # type: ignore
          repr: bool = True, hash: Optional[bool] = None, compare: bool = True,
          metadata: Optional[Mapping[str, Any]] = None) -> Any:
    """
    Args:
        static: Indicates whether a field is a pytree or static.  Pytree fields are
            differentiated and traced.  Static fields are hashed and compared.
    """
    if metadata is None:
        metadata = {}
    return dataclasses.field(metadata={**metadata, 'static': static},
                             default=default, default_factory=default_factory, init=init, repr=repr,
                             hash=hash, compare=compare)  # type: ignore


def fields(d: Any, *, static: Optional[bool] = None) -> Iterable[Field[Any]]:
    if static is None:
        yield from d_fields(d)
    for this_field in d_fields(d):
        if this_field.metadata.get('static', False) == static:
            yield this_field


def field_names(d: Any, *, static: Optional[bool] = None) -> Iterable[str]:
    for this_field in fields(d, static=static):
        yield this_field.name


def field_names_and_values(d: Any, *, static: Optional[bool] = None) -> Iterable[Tuple[str, Any]]:
    for name in field_names(d, static=static):
        yield name, getattr(d, name)


def field_values(d: Any, *, static: Optional[bool] = None) -> Iterable[Any]:
    for name in field_names(d, static=static):
        yield getattr(d, name)


def field_names_values_metadata(d: Any, *, static: Optional[bool] = None) -> (
        Iterable[Tuple[str, Any, Mapping[str, Any]]]):
    for this_field in fields(d, static=static):
        yield this_field.name, getattr(d, this_field.name), this_field.metadata


def document_dataclass(pdoc: MutableMapping[str, Any], name: str) -> None:
    pdoc[f'{name}.static_fields'] = False
    pdoc[f'{name}.nonstatic_fields'] = False
    pdoc[f'{name}.tree_flatten'] = False
    pdoc[f'{name}.tree_unflatten'] = False
    pdoc[f'{name}.display'] = False
    pdoc[f'{name}.replace'] = False



from __future__ import annotations

from typing import Generic, TypeVar

from ..annotations import PyTree
from ..dataclass import dataclass

__all__ = ['AugmentedState']


State = TypeVar('State', bound=PyTree)


@dataclass
class AugmentedState(Generic[State]):

    current_state: State
    iterations: int



from typing import Any

from .chain import ChainedGradientTransformation
from .transform import GradientTransformation
from .transforms import AdditiveWeightDecay, Scale, ScaleByAdam

__all__ = ['adam', 'adamw']


def adam(learning_rate: float,
         beta1: float = 0.9,
         beta2: float = 0.999,
         epsilon: float = 1e-8) -> GradientTransformation[Any, Any]:
    return ChainedGradientTransformation([ScaleByAdam(beta1=beta1, beta2=beta2, epsilon=epsilon),
                                          Scale(-learning_rate)])


def adamw(learning_rate: float,
          beta1: float = 0.9,
          beta2: float = 0.999,
          epsilon: float = 1e-8,
          weight_decay: float = 1e-4) -> GradientTransformation[Any, Any]:
    return ChainedGradientTransformation([ScaleByAdam(beta1=beta1, beta2=beta2, epsilon=epsilon),
                                          AdditiveWeightDecay(weight_decay),
                                          Scale(-learning_rate)])



from typing import Callable, Generic, Optional, Tuple, TypeVar

from chex import Array, Numeric
from jax import numpy as jnp
from jax.tree_util import tree_map, tree_multimap

from ..annotations import PyTree
from ..dataclass import dataclass
from .transform import SecondOrderGradientTransformation

__all__ = ['SMDState', 'SMDGradient']


Weights = TypeVar('Weights', bound=PyTree)


@dataclass
class SMDState(Generic[Weights]):
    log_learning_rate: Weights
    v: Weights


@dataclass
class SMDGradient(SecondOrderGradientTransformation[SMDState[Weights], Weights],
                  Generic[Weights]):
    """
    Schraudolph, N. N. (1999). Local gain adaptation in stochastic gradient descent. Artificial
    Neural Networks, 1999. ICANN 99. Ninth International Conference on (Conf. Publ. No. 470), 2,
    569–574. https://doi.org/10.1049/cp:19991170
    """
    meta_learning_rate: Numeric = 1e-2

    def init(self, parameters: Weights) -> SMDState[Weights]:
        z = tree_map(jnp.zeros_like, parameters)
        return SMDState[Weights](z, z)

    def second_order_update(self,
                            gradient: Weights,
                            state: SMDState[Weights],
                            parameters: Optional[Weights],
                            hessian_vector_product: Callable[[Weights], Weights]) -> (
                                Tuple[Weights, SMDState[Weights]]):
        negative_gradient = tree_map(jnp.negative, gradient)  # delta

        # Update log-learning rate.
        def g(log_p: Array, delta: Array, v: Array) -> Array:
            return log_p + self.meta_learning_rate * delta * v

        new_log_learning_rate = tree_multimap(g, state.log_learning_rate, negative_gradient,
                                              state.v)
        learning_rate = tree_map(jnp.exp, new_log_learning_rate)  # p

        # Calculate gradient.
        gradient = tree_multimap(jnp.multiply, learning_rate, negative_gradient)

        # Update v.
        def f(v: Array, p: Array, delta: Array, hv: Array) -> Array:
            return v + p * delta - hv

        new_v = tree_multimap(f, state.v, learning_rate, negative_gradient,
                              hessian_vector_product(state.v))

        return gradient, SMDState[Weights](new_log_learning_rate, new_v)



# This patch changes dataclass behavior for all dataclasses:  When a non-defaulted dataclass field
# follows a defaulted field, the behaviour has changed from raising a TypeError to making all
# arguments in __init__ following the defaulted field keyword-only.
# See: https://github.com/python/cpython/pull/17322
# type: ignore
# pylint: disable=protected-access
import dataclasses
import functools as ft
import sys

__all__ = []


@ft.wraps(dataclasses._init_fn)
def _init_fn(fields, frozen, has_post_init, self_name, globals_=None):
    """Build ``__init__`` for a data-class."""
    py_37 = sys.version_info < (3, 7, 6)

    locals_ = {f"_type_{f.name}": f.type for f in fields}
    extra = {
        "MISSING": dataclasses.MISSING,
        "_HAS_DEFAULT_FACTORY": dataclasses._HAS_DEFAULT_FACTORY,
    }
    if py_37:
        assert globals_ is None
        globals_ = extra
    else:
        assert globals_ is not None
        locals_.update(extra)

    body_lines = []
    for f in fields:
        line = dataclasses._field_init(
            f, frozen, globals_ if py_37 else locals_, self_name
        )
        if line:
            body_lines.append(line)
    if has_post_init:
        params_str = ",".join(
            f.name for f in fields if f._field_type is dataclasses._FIELD_INITVAR
        )
        body_line = f"{self_name}.{dataclasses._POST_INIT_NAME}({params_str})"
        body_lines.append(body_line)
    if not body_lines:
        body_lines = ["pass"]

    # Edit: args after defaulted args are keyword-only
    seen_default = False
    keyword_only = False
    args = [self_name]
    for f in fields:
        if f.init:
            has_default = f.default is not dataclasses.MISSING
            has_default_factory = f.default_factory is not dataclasses.MISSING
            if has_default or has_default_factory:
                seen_default = True
            elif seen_default and not keyword_only:
                keyword_only = True
                args.append("*")
            args.append(dataclasses._init_param(f))

    return dataclasses._create_fn(
        "__init__",
        args,
        body_lines,
        locals=locals_,
        globals=globals_,
        return_type=None,
    )


dataclasses._init_fn = _init_fn



from __future__ import annotations

import sys
from typing import Hashable, Sequence, Tuple, Type, TypeVar

from .annotations import PyTree

__all__ = ['PyTreeLike']


T = TypeVar('T', bound='PyTreeLike')


if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable

    @runtime_checkable
    class PyTreeLike(Protocol):

        def tree_flatten(self) -> Tuple[Sequence[PyTree], Hashable]:
            """
            Returns:
                trees: A JAX PyTree of trees representing the object.
                hashed: Data that will be treated as constant through JAX operations.
            """
            ...

        @classmethod
        def tree_unflatten(cls: Type[T], hashed: Hashable, trees: Sequence[PyTree]) -> T:
            """
            Args:
                hashed: Data that will be treated as constant through JAX operations.
                trees: A JAX PyTree of trees from which the object is constructed.
            Returns:
                A constructed object.
            """
            ...
else:
    class PyTreeLike:
        pass



from __future__ import annotations

from typing import Any, Callable, Dict, Sequence, Tuple, TypeVar, Union

from chex import Array
from jax import vjp

__all__ = ['Shape',
           'ShapeLike',
           'SliceLike',
           'Array',
           'RealArray',
           'ComplexArray',
           'PyTree',
           'TapFunctionTransforms',
           'vjp_with_aux']


Shape = Tuple[int, ...]
ShapeLike = Union[int, Sequence[int]]
SliceLike = Tuple[Union[int, None, slice], ...]
RealArray = Array
ComplexArray = Array
PyTree = Any
TapFunctionTransforms = Sequence[Tuple[str, Dict[str, Any]]]


T = TypeVar('T')
U = TypeVar('U')


def vjp_with_aux(fun: Callable[..., Tuple[T, U]],
                 *primals: Any) -> Tuple[T, Callable[..., Any], U]:
    return vjp(fun, *primals, has_aux=True)  # type: ignore



from __future__ import annotations

from typing import Generic, Optional, Sequence, TypeVar, Union

import numpy as np
from chex import Array
from jax.tree_util import tree_map
from matplotlib.axes import Axes

from .annotations import PyTree
from .dataclass import dataclass
from .leaky_integral import leaky_integrate_time_series

__all__ = ['PlottableTrajectory']


Trajectory = TypeVar('Trajectory', bound=PyTree)


@dataclass
class PlottableTrajectory(Generic[Trajectory]):
    trajectory: Trajectory
    "The trajectory is a PyTree containing the plotted data in its nonstatic attributes."
    times: np.ndarray
    "The times corresponding to the data points in each of the plotted attributes."

    def plot(self,
             data: Array,
             axis: Axes,
             *,
             title: Optional[str] = None,
             legend: int = 0,
             labels: Optional[Sequence[str]] = None,
             decay: Optional[float] = None,
             clip_slice: slice = slice(None)) -> None:
        """
        Plot the PlottableTrajectory into a matplotlib axis.
        Args:
            data: The data to be plotted.  Typically, this is a sub-element of trajectory.
            axis: A matplotlib axis to plot into.
            title: The title of the subplot.
            legend: The maximum number of entries for which to generate a legend.
            decay: If not none, applies a leaky-integral with the given decay to the plotted points.
            clip_slice: Restrict the plot to a slice.
        """
        if title is not None:
            axis.set_title(title)
        axis.ticklabel_format(useOffset=False)

        times = self.times[clip_slice]
        all_ys = np.asarray(data)
        if not np.all(np.isfinite(all_ys)):
            print(f"Graph {'' if title is None else title + ' '}contains infinite numbers.")
            all_ys = np.where(np.isfinite(all_ys), all_ys, 0.0)

        plot_indices = list(np.ndindex(all_ys.shape[1:]))
        if labels is None:
            labels = [str(x) for x in plot_indices]
        for plot_index, label in zip(plot_indices, labels):
            ys = all_ys[(clip_slice, *plot_index)]
            if decay is not None:
                ys = leaky_integrate_time_series(ys, decay)
            axis.plot(times, ys, label=label)
        number_of_graph_lines = np.prod(data.shape[1:])
        if labels is not None or legend > 0 and 0 < number_of_graph_lines <= legend:
            axis.legend()

    def slice_into(self, s: Union[int, slice]) -> Trajectory:
        "Return a new trajectory whose nonstatic attributed are sliced."
        return tree_map(lambda x: x[s], self.trajectory)



from __future__ import annotations

from functools import partial
from typing import Generic, Tuple

from chex import Array
from jax import numpy as jnp
from jax.tree_util import tree_map, tree_multimap, tree_reduce

from ..dataclass import dataclass
from ..leaky_integral import leaky_data_weight, leaky_integrate
from ..tools import abs_square
from .augmented import AugmentedState, State
from .combinator import Differentiand, IteratedFunctionWithCombinator
from .iterated_function import Comparand, IteratedFunction, Parameters, Trajectory

__all__ = ['StochasticState', 'StochasticIteratedFunction',
           'StochasticIteratedFunctionWithCombinator']


@dataclass
class StochasticState(AugmentedState[State], Generic[State, Comparand]):

    mean_state: Comparand
    second_moment_state: Comparand


# https://github.com/python/mypy/issues/8539
@dataclass  # type: ignore
class StochasticIteratedFunction(
        IteratedFunction[Parameters, State, Comparand, Trajectory,
                         StochasticState[State, Comparand]],
        Generic[Parameters, State, Comparand, Trajectory]):

    convergence_detection_decay: float

    # Implemented methods --------------------------------------------------------------------------
    def initial_augmented(self, initial_state: State) -> StochasticState[State, Comparand]:
        mean_state, second_moment_state = self._sufficient_statistics(initial_state)
        return StochasticState(current_state=initial_state,
                               iterations=0,
                               mean_state=mean_state,
                               second_moment_state=second_moment_state)

    def iterate_augmented(self,
                          new_state: State,
                          augmented: StochasticState[State, Comparand]) -> (
                              StochasticState[State, Comparand]):
        def f(value: Array, drift: Array) -> Array:
            return leaky_integrate(value, 1.0, drift, self.convergence_detection_decay,
                                   leaky_average=True)

        mean_state, second_moment_state = self._sufficient_statistics(augmented.current_state)
        new_mean_state = tree_multimap(f, augmented.mean_state, mean_state)
        new_second_moment_state = tree_multimap(f, augmented.second_moment_state,
                                                second_moment_state)
        return StochasticState(current_state=new_state,
                               iterations=augmented.iterations + 1,
                               mean_state=new_mean_state,
                               second_moment_state=new_second_moment_state)

    def converged(self, augmented: StochasticState[State, Comparand]) -> Array:
        data_weight = leaky_data_weight(augmented.iterations, self.convergence_detection_decay)
        mean_squared = tree_map(abs_square, augmented.mean_state)
        return tree_reduce(jnp.logical_and,
                           tree_multimap(partial(jnp.allclose,
                                                 rtol=self.rtol * data_weight,
                                                 atol=self.atol * data_weight),
                                         augmented.second_moment_state,
                                         mean_squared),
                           True)

    # New methods ----------------------------------------------------------------------------------
    def minimum_tolerances(self, augmented: StochasticState[State, Comparand]) -> Tuple[Array,
                                                                                        Array]:
        """
        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        def safe_divide(numerator: Array, denominator: Array) -> Array:
            return jnp.where(denominator > 0.0, numerator / denominator, jnp.inf)

        data_weight = leaky_data_weight(augmented.iterations, self.convergence_detection_decay)
        mean_squared = tree_map(abs_square, augmented.mean_state)
        variance = tree_multimap(jnp.subtract, augmented.second_moment_state, mean_squared)
        scaled_variance = tree_multimap(safe_divide, variance, mean_squared)

        minium_atol = safe_divide(tree_reduce(jnp.maximum, tree_map(jnp.amax, variance)),
                                  data_weight)
        minium_rtol = safe_divide(tree_reduce(jnp.maximum, tree_map(jnp.amax, scaled_variance)),
                                  data_weight)
        return minium_atol, minium_rtol

    # Private methods ------------------------------------------------------------------------------
    def _sufficient_statistics(self, state: State) -> Tuple[Comparand, Comparand]:
        comparand = self.extract_comparand(state)
        squared_comparand = tree_map(abs_square, comparand)
        return comparand, squared_comparand


class StochasticIteratedFunctionWithCombinator(
        IteratedFunctionWithCombinator[Parameters, State, Comparand, Differentiand, Trajectory,
                                       StochasticState[State, Comparand]],
        StochasticIteratedFunction[Parameters, State, Comparand, Trajectory],
        Generic[Parameters, State, Comparand, Differentiand, Trajectory]):
    pass



from functools import partial, singledispatch
from numbers import Complex, Number, Real
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
from jax import numpy as jnp
from jax.interpreters.xla import DeviceArray
from jax.tree_util import tree_multimap, tree_reduce

from .annotations import PyTree
from .dtypes import default_atol, default_rtol

__all__ = ['assert_jax_allclose', 'jax_allclose', 'get_test_string', 'get_relative_test_string']


def assert_jax_allclose(actual: PyTree,
                        desired: PyTree,
                        original_name: Optional[str] = None,
                        original_value: Optional[PyTree] = None,
                        *,
                        rtol: Optional[float] = None,
                        atol: Optional[float] = None) -> None:
    """
    Asserts that every tensor in an actual pytree matches the corresponding tensor in a desired
    pytree.  If the assertion fails, a passing test string is printed::

    ```python
    from tjax import assert_jax_allclose, dataclass, Array

    @dataclass
    class A:
        x: Array
        y: Array

    @dataclass
    class B:
        z: A

    original = B(A(1.2, 3.4))
    desired = B(A(3.0, 4.0))
    actual = B(A(1.2, 5.2))

    assert_jax_allclose(actual, desired, 'original', original)
    ```
    This prints::
    ```
    JAX trees don't match.  Actual:
    B
        z=A
            x=3.0
            y=4.0

    Desired:
    B
        z=A
            x=1.2
            y=5.2

    Test string:
    original.replace(z=original.z.replace(x=3.0, y=4.0))
    ```
    The test string can then be pasted.

    Args:
        actual: The actual value.
        desired: The desired value.
        original_name: The variable name that contains the original value.
        original_value: The original value.  This is usually a pytree like a dataclass that has the
            same type as actual and desired, but contains different values.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    """
    if rtol is None:
        rtol = default_rtol()
    if atol is None:
        atol = default_atol()

    try:
        tree_multimap(partial(np.testing.assert_allclose, rtol=rtol, atol=atol), actual, desired)
    except AssertionError:
        print("JAX trees don't match.\nActual:")
        print(actual)
        print("Desired:")
        print(desired)
        print("Test string:")
        if original_name is not None and original_value is not None:
            print(get_relative_test_string(actual, original_name, original_value, rtol, atol))
        else:
            print(get_test_string(actual, rtol, atol))
        raise


def jax_allclose(actual: PyTree,
                 desired: PyTree,
                 rtol: Optional[float] = None,
                 atol: Optional[float] = None) -> bool:
    """
    Args:
        actual: The actual value.
        desired: The desired value.
        rtol: The relative tolerance of the comparisons in the comparison.
        atol: The absolute tolerance of the comparisons in the comparison.
    """
    if rtol is None:
        rtol = default_rtol()
    if atol is None:
        atol = default_atol()

    return cast(
        bool,
        tree_reduce(jnp.logical_and,
                    tree_multimap(partial(np.allclose, rtol=rtol, atol=atol), actual, desired),
                    True))


# get test string ----------------------------------------------------------------------------------
# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/2904.
@singledispatch
def get_test_string(actual: Any, rtol: float, atol: float) -> str:
    """
    Args:
        actual: The actual value that was produced, and that should be the desired value.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    Returns:
        A string of Python code that produces the desired value.
    """
    return str(actual)


@get_test_string.register(np.ndarray)
@get_test_string.register(DeviceArray)
def _(actual: Union[np.ndarray, DeviceArray], rtol: float, atol: float) -> str:
    def fts(x: Complex) -> str:
        return _float_to_string(x, rtol, atol)
    with np.printoptions(formatter={'float_kind': fts,
                                    'complex_kind': fts}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace(
            '  ', ' ')


@get_test_string.register  # type: ignore
def _(actual: Complex, rtol: float, atol: float) -> str:
    return _float_to_string(complex(actual), rtol, atol)


@get_test_string.register  # type: ignore
def _(actual: Number, rtol: float, atol: float) -> str:
    return str(actual)


@get_test_string.register(list)  # type: ignore
@get_test_string.register(tuple)
def _(actual: Union[List[Any], Tuple[Any]], rtol: float, atol: float) -> str:
    is_list = isinstance(actual, list)
    return (("[" if is_list else "(")
            + ", ".join(get_test_string(sub_actual, rtol, atol)
                        for i, sub_actual in enumerate(actual))
            + ("]" if is_list else ")"))


@get_test_string.register(dict)  # type: ignore
def _(actual: Dict[Any, Any], rtol: float, atol: float) -> str:
    return '{' + ",\n".join(repr(key) + ': ' + get_test_string(sub_actual, rtol, atol)
                            for key, sub_actual in actual.items()) + '}'


# get relative test string -------------------------------------------------------------------------
@singledispatch
def get_relative_test_string(actual: Any,
                             original_name: str,
                             original: Any,
                             rtol: float,
                             atol: float) -> str:
    """
    Args:
        original_name: The name of the variable containing an original value.
        actual: The actual value that was produced, and that should be the desired value.
        original: The original value.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    Returns:
        A string of Python code that produces the desired value from an "original value" (could be
        zeroed-out, for example).
    """
    return str(actual)


@get_relative_test_string.register(np.ndarray)  # type: ignore
@get_relative_test_string.register(DeviceArray)
def _(actual: Union[np.ndarray, DeviceArray], original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    def fts(x: Complex) -> str:
        return _float_to_string(x, rtol, atol)
    with np.printoptions(formatter={'float_kind': fts,
                                    'complex_kind': fts}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace(
            '  ', ' ')


@get_relative_test_string.register  # type: ignore
def _(actual: Complex, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return _float_to_string(actual, rtol, atol)


@get_relative_test_string.register  # type: ignore
def _(actual: Number, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return str(actual)


@get_relative_test_string.register(list)  # type: ignore
@get_relative_test_string.register(tuple)
def _(actual: Union[List[Any], Tuple[Any]], original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    is_list = isinstance(actual, list)
    return (("[" if is_list else "(")
            + ", ".join(get_relative_test_string(f"{original_name}[{i}]",
                                                 sub_actual, sub_original,
                                                 rtol, atol)
                        for i, (sub_actual, sub_original) in enumerate(zip(actual, original)))
            + ("]" if is_list else ")"))


@get_relative_test_string.register(dict)  # type: ignore
def _(actual: Dict[Any, Any], original_name: str, original: Any, rtol: float, atol: float) -> str:
    if not isinstance(original, dict):
        raise TypeError

    def relative_string(key, sub_actual):
        return get_relative_test_string(
            f"{original_name}[{key}]", sub_actual, original[key], rtol, atol)

    return '{' + ",\n".join(repr(key) + ': ' + relative_string(key, sub_actual)
                            for key, sub_actual in actual.items()) + '}'


# Private functions --------------------------------------------------------------------------------
def _float_to_string_with_precision(x: Union[float, complex], precision: int) -> str:
    with np.printoptions(precision=precision, floatmode='maxprec'):
        return repr(np.array(x))[6:-1]


def _float_to_string(x: complex, rtol: float, atol: float) -> str:
    y: Union[float, complex]
    if isinstance(x, Real):
        y = float(x)
    elif isinstance(x, Complex):
        y = complex(x)
    else:
        raise TypeError
    for i in range(20):
        retval = _float_to_string_with_precision(y, i)
        if np.allclose(complex(retval), x, rtol=rtol, atol=atol):
            break
    return retval



from typing import Generic, Optional, Tuple

from chex import Numeric
from optax import ScaleByAdamState, ScaleState, additive_weight_decay, scale, scale_by_adam

from ..dataclass import dataclass
from .transform import GradientTransformation, Weights

__all__ = ['Scale', 'ScaleByAdam', 'AdditiveWeightDecay']


@dataclass
class Scale(GradientTransformation[ScaleState, Weights], Generic[Weights]):
    step_size: Numeric

    def init(self, parameters: Weights) -> ScaleState:
        return scale(self.step_size).init(parameters)

    def update(self,
               gradient: Weights,
               state: ScaleState,
               parameters: Optional[Weights]) -> Tuple[Weights, ScaleState]:
        return scale(self.step_size).update(gradient, state, parameters)


@dataclass
class ScaleByAdam(GradientTransformation[ScaleByAdamState, Weights], Generic[Weights]):
    beta1: Numeric = 0.9
    beta2: Numeric = 0.999
    epsilon: Numeric = 1e-8
    epsilon_root: Numeric = 0.0

    def init(self, parameters: Weights) -> ScaleByAdamState:
        return scale_by_adam(self.beta1, self.beta2, self.epsilon, self.epsilon_root).init(
            parameters)

    def update(self,
               gradient: Weights,
               state: ScaleByAdamState,
               parameters: Optional[Weights]) -> Tuple[Weights, ScaleByAdamState]:
        return scale_by_adam(self.beta1, self.beta2, self.epsilon, self.epsilon_root).update(
            gradient, state, parameters)


@dataclass
class AdditiveWeightDecay(GradientTransformation[None, Weights], Generic[Weights]):
    weight_decay: Numeric = 0.0

    def init(self, parameters: Weights) -> None:
        return None

    def update(self,
               gradient: Weights,
               state: None,
               parameters: Optional[Weights]) -> Tuple[Weights, ScaleState]:
        return additive_weight_decay(self.weight_decay).update(gradient, state, parameters)



from __future__ import annotations

from functools import partial
from typing import Generic

from chex import Array
from jax import numpy as jnp
from jax.tree_util import tree_multimap, tree_reduce

from ..dataclass import dataclass
from .augmented import AugmentedState, State
from .iterated_function import Comparand, IteratedFunction, Parameters, Trajectory

__all__ = ['ComparingState', 'ComparingIteratedFunction']


@dataclass
class ComparingState(AugmentedState[State], Generic[State, Comparand]):

    last_state: Comparand


# https://github.com/python/mypy/issues/8539
@dataclass  # type: ignore
class ComparingIteratedFunction(
        IteratedFunction[Parameters, State, Comparand, Trajectory,
                         ComparingState[State, Comparand]],
        Generic[Parameters, State, Comparand, Trajectory]):

    # Implemented methods --------------------------------------------------------------------------
    def initial_augmented(self, initial_state: State) -> ComparingState[State, Comparand]:
        return ComparingState(current_state=initial_state,
                              iterations=0,
                              last_state=self.extract_comparand(initial_state))

    def expected_state(self, theta: Parameters, state: State) -> State:
        return self.sampled_state(theta, state)

    def iterate_augmented(self,
                          new_state: State,
                          augmented: ComparingState[State, Comparand]) -> (
                              ComparingState[State, Comparand]):
        return ComparingState(current_state=new_state,
                              iterations=augmented.iterations + 1,
                              last_state=self.extract_comparand(augmented.current_state))

    def converged(self, augmented: ComparingState[State, Comparand]) -> Array:
        return tree_reduce(jnp.logical_and,
                           tree_multimap(partial(jnp.allclose, rtol=self.rtol, atol=self.atol),
                                         self.extract_comparand(augmented.current_state),
                                         augmented.last_state))
